from abc import ABC, abstractmethod, abstractproperty


class PitopCliException(Exception):
    pass


class PitopCliInvalidArgument(Exception):
    pass


class CliBaseClass(ABC):
    """Abstract class, used to create CLI commands."""

    @abstractmethod
    def __init__(self, args):
        """Class constructor.

        Args:
            args (Namespace): arguments as returned by Argparse.parse_args
        """
        pass

    @abstractmethod
    def run(self):
        """Executes the action performed by the CLI."""
        pass

    @classmethod
    @abstractmethod
    def add_parser_arguments(cls, parser):
        """Add CLI expected arguments to the provided parser.

        Args:
            parser (argparse._SubParsersAction | argparse.ArgumentParser): parser where class arguments will be appended
        """
        pass

    @property
    @abstractproperty
    def parser_help(self):
        """Help string to be displayed by ArgumentParser."""
        pass

    @property
    @abstractproperty
    def cli_name(self):
        """Name of the class CLI, without the 'pt-' prefix."""
        pass

    @property
    def parser(self):
        """ArgumentParser object used to parse the class."""
        pass

    def validate_args(self):
        """Method to perform further validation on arguments."""
        pass



class TypeHelper:
    @staticmethod
    def is_integer(string):
        try:
            int(string)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_float(string):
        try:
            float(string)
            return True
        except ValueError:
            return False



from concurrent.futures import ThreadPoolExecutor
from inspect import signature

from numpy import asarray

from pitop.processing.core.vision_functions import import_opencv

from .capture_action_base import CaptureActionBase


class MotionDetector(CaptureActionBase):
    """Class that implements a motion detection algorithm. When motion is
    detected, a callback function is executed.

    :param function callback_on_motion: A callback function that will be called when motion is detected.
    :param int moving_object_minimum_area: The sensitivity of the motion detection, measured as the area of
    pixels changing between frames that constitutes motion.
    """

    def __init__(self, callback_on_motion, moving_object_minimum_area):
        self.cv2 = import_opencv()

        self.__motion_detect_previous_frame = None
        self.__motion_detect_threshold = moving_object_minimum_area ** 2
        self.__motion_detect_callback = callback_on_motion
        self.__event_executor = ThreadPoolExecutor()

        callback_signature = signature(callback_on_motion)
        self.callback_has_argument = len(callback_signature.parameters) > 0

    def __del__(self):
        self.stop()

    def process(self, frame):
        frame = asarray(frame)
        # Use greyscale and blurred for motion detection
        gray = self.cv2.cvtColor(frame, self.cv2.COLOR_BGR2GRAY)
        gray = self.cv2.GaussianBlur(gray, (21, 21), 0)

        if self.__motion_detect_previous_frame is None:
            self.__motion_detect_previous_frame = gray
        else:
            diff_frame = self.cv2.absdiff(self.__motion_detect_previous_frame, gray)
            threshold_frame = self.cv2.threshold(
                diff_frame, 30, 255, self.cv2.THRESH_BINARY
            )[1]
            threshold_frame = self.cv2.dilate(threshold_frame, None, iterations=2)

            contour_data = self.cv2.findContours(
                threshold_frame.copy(),
                self.cv2.RETR_EXTERNAL,
                self.cv2.CHAIN_APPROX_SIMPLE,
            )
            if len(contour_data) == 2:
                # opencv 2
                detected_contours = contour_data[0]
            elif len(contour_data) == 3:
                # opencv 3
                detected_contours = contour_data[1]

            for contour in detected_contours:
                area = self.cv2.contourArea(contour)

                if area > self.__motion_detect_threshold:
                    if self.callback_has_argument:
                        self.__event_executor.submit(
                            self.__motion_detect_callback, frame
                        )
                    else:
                        self.__event_executor.submit(self.__motion_detect_callback)
                    break

    def stop(self):
        if self.__event_executor is not None:
            self.__event_executor.shutdown()
            self.__event_executor = None



from .cli_base import CliBaseClass


class BatteryCLI(CliBaseClass):
    parser_help = "Get battery information from a pi-top"
    cli_name = "battery"

    def __init__(self, args) -> None:
        self.args = args
        self.args_order = list()

    def run(self) -> int:
        try:
            self.print_battery_state()
            return 0
        except Exception as e:
            print(f"Error on pitop-battery.run: {e}")
            return 1

    def print_battery_state(self) -> None:
        if self.args.charging_state:
            self.args_order.append("charging-state")
        if self.args.capacity:
            self.args_order.append("capacity")
        if self.args.time_remaining:
            self.args_order.append("time-remaining")
        if self.args.wattage:
            self.args_order.append("wattage")

        from pitop.battery import Battery

        charging_state, capacity, time_remaining, wattage = Battery.get_full_state()

        if len(self.args_order) > 0:
            for arg in set(self.args_order):
                if arg == "charging-state":
                    print(charging_state)
                if arg == "capacity":
                    print(capacity)
                if arg == "time-remaining":
                    print(time_remaining)
                if arg == "wattage":
                    print(wattage)
        else:

            if self.args.charging_state:
                print(charging_state)
            elif self.args.capacity:
                print(capacity)
            elif self.args.time_remaining:
                print(time_remaining)
            elif self.args.wattage:
                print(wattage)
            else:
                print("Charging State: " + charging_state)
                print("Capacity: " + capacity)
                print("Time Remaining: " + time_remaining)
                print("Wattage: " + wattage)

    @classmethod
    def add_parser_arguments(cls, parser):
        parser.add_argument(
            "-s",
            "--charging-state",
            help="Get charging state. -1 = No pi-top battery detected, 0 = Discharging, 1 = Charging, 2 = Full battery",
            action="store_true",
        )
        parser.add_argument(
            "-c",
            "--capacity",
            help="Get battery capacity percentage (%%)",
            action="store_true",
        )
        parser.add_argument(
            "-t",
            "--time-remaining",
            help="Get the time (in minutes) to full or time to empty based on the charging state",
            action="store_true",
        )
        parser.add_argument(
            "-w",
            "--wattage",
            help="Get the wattage (mAh) of the battery",
            action="store_true",
        )
        parser.add_argument("-v", "--verbose", action="count")


def main():
    from .deprecated_cli_runner import run

    run(BatteryCLI)


if __name__ == "__main__":
    main()



from time import sleep

try:
    from time import monotonic
except ImportError:
    from monotonic import monotonic


class FPS_Regulator(object):
    """Adapted from ``luma.core.spritesheet``

    Implements a variable sleep mechanism to give the appearance of a consistent
    frame rate. Using a fixed-time sleep will cause animations to be jittery
    (looking like they are speeding up or slowing down, depending on what other
    work is occurring), whereas this class keeps track of when the last time the
    ``sleep()`` method was called, and calculates a sleep period to smooth out
    the jitter.
    :param fps: The max frame rate, expressed numerically in
        frames-per-second.  By default, this is set at 16.67, to give a frame
        render time of approximately 60ms. This can be overridden as necessary,
        and if no FPS limiting is required, the ``fps`` can be set to zero.
    :type fps: float
    """

    def __init__(self, max_fps=16.67):
        self.set_max_fps(max_fps)
        self.total_transit_time = 0
        self.start_time = None
        self.last_time = None
        self.fps = 30

    def set_max_fps(self, max_fps):
        """Sets the max frame rate to the value provided.

        Parameters
        ----------
        max_fps : int
          The frame rate that the user aims to render to the screen at. Lower frame rates are permitted.
        """
        self.total_no_of_frames = 0
        if max_fps == 0:
            max_fps = -1
        self.max_sleep_time = 1.0 / max_fps

    def throttle_fps_if_needed(self):
        """Sleep to ensure that max frame rate is not exceeded."""
        if self.max_sleep_time >= 0:
            last_frame_transit_time = monotonic() - self.last_time
            sleep_for = self.max_sleep_time - last_frame_transit_time

            if sleep_for > 0:
                sleep(sleep_for)

    def start_timer(self):
        """Starts internal timer so that time taken to render frame can be
        known."""
        self.enter_time = monotonic()
        if not self.start_time:
            self.start_time = self.enter_time
            self.last_time = self.enter_time

    def stop_timer(self):
        """Stops internal timer so that time taken to render frame can be
        known.

        Responsible for throttling frame rate as required
        """
        try:
            self.total_transit_time += monotonic() - self.enter_time
            self.total_no_of_frames += 1
            self.throttle_fps_if_needed()

            self.last_time = monotonic()
        except AttributeError:
            pass

    def effective_FPS(self):
        """
        Calculates the effective frames-per-second - this should largely
        correlate to the max FPS supplied in the constructor, but no
        guarantees are given.
        :returns: the effective frame rate
        :rtype: float
        """
        if self.start_time is None:
            self.start_time = 0
        elapsed = monotonic() - self.start_time
        return self.total_no_of_frames / elapsed

    def average_transit_time(self):
        """
        Calculates the average transit time between the enter and exit methods,
        and return the time in milliseconds
        :returns: the average transit in milliseconds
        :rtype: float
        """
        return self.total_transit_time * 1000.0 / self.total_no_of_frames



import atexit

from pitop.common.ptdm import Message, PTDMSubscribeClient

from .oled import OLED


class Miniscreen(OLED):
    """Represents a pi-top [4]'s miniscreen display.

    Also owns the surrounding 4 buttons as properties
    (:class:`up_button`, :class:`down_button`, :class:`select_button`,
    :class:`cancel_button`). See
    :class:`pitop.miniscreen.miniscreen.MiniscreenButton` for how to use these
    buttons.
    """

    def __init__(self):
        super(Miniscreen, self).__init__()

        self._up_button = MiniscreenButton()
        self._down_button = MiniscreenButton()
        self._select_button = MiniscreenButton()
        self._cancel_button = MiniscreenButton()

        self.__ptdm_subscribe_client = None
        self.__setup_subscribe_client()

        atexit.register(self.__clean_up)

    def __setup_subscribe_client(self):
        def set_button_state(button, pressed):
            button.is_pressed = pressed
            self.__ptdm_subscribe_client.invoke_callback_func_if_exists(
                button.when_pressed if button.is_pressed else button.when_released
            )

        self.__ptdm_subscribe_client = PTDMSubscribeClient()
        self.__ptdm_subscribe_client.initialise(
            {
                Message.PUB_V3_BUTTON_UP_PRESSED: lambda: set_button_state(
                    self._up_button, pressed=True
                ),
                Message.PUB_V3_BUTTON_UP_RELEASED: lambda: set_button_state(
                    self._up_button, pressed=False
                ),
                # ----------------------------------------------------------------------------------
                Message.PUB_V3_BUTTON_DOWN_PRESSED: lambda: set_button_state(
                    self._down_button, pressed=True
                ),
                Message.PUB_V3_BUTTON_DOWN_RELEASED: lambda: set_button_state(
                    self._down_button, pressed=False
                ),
                # ----------------------------------------------------------------------------------
                Message.PUB_V3_BUTTON_SELECT_PRESSED: lambda: set_button_state(
                    self._select_button, pressed=True
                ),
                Message.PUB_V3_BUTTON_SELECT_RELEASED: lambda: set_button_state(
                    self._select_button, pressed=False
                ),
                # ----------------------------------------------------------------------------------
                Message.PUB_V3_BUTTON_CANCEL_PRESSED: lambda: set_button_state(
                    self._cancel_button, pressed=True
                ),
                Message.PUB_V3_BUTTON_CANCEL_RELEASED: lambda: set_button_state(
                    self._cancel_button, pressed=False
                ),
            }
        )
        self.__ptdm_subscribe_client.start_listening()

    def __clean_up(self):
        try:
            self.__ptdm_subscribe_client.stop_listening()
        except Exception:
            pass

    @property
    def up_button(self):
        """Gets the up button of the pi-top [4] miniscreen.

        :return: A gpiozero-like button instance representing the up button of the pi-top [4] miniscreen.
        :rtype: :class:`pitop.miniscreen.miniscreen.MiniscreenButton`
        """
        return self._up_button

    @property
    def down_button(self):
        """Gets the down button of the pi-top [4] miniscreen.

        :return: A gpiozero-like button instance representing the down button of the pi-top [4] miniscreen.
        :rtype: :class:`pitop.miniscreen.miniscreen.MiniscreenButton`
        """
        return self._down_button

    @property
    def select_button(self):
        """Gets the select button of the pi-top [4] miniscreen.

        :return: A gpiozero-like button instance representing the select button of the pi-top [4] miniscreen.
        :rtype: :class:`pitop.miniscreen.miniscreen.MiniscreenButton`
        """
        return self._select_button

    @property
    def cancel_button(self):
        """Gets the cancel button of the pi-top [4] miniscreen.

        :return: A gpiozero-like button instance representing the cancel button of the pi-top [4] miniscreen.
        :rtype: :class:`pitop.miniscreen.miniscreen.MiniscreenButton`
        """
        return self._cancel_button


class MiniscreenButton:
    """Represents one of the 4 buttons around the miniscreen display on a pi-
    top [4].

    Should not be created directly - instead, use :class:`pitop.miniscreen.Miniscreen`.
    """

    def __init__(self):
        # State parameters
        self._is_pressed = False
        # Event-based functions
        self._when_pressed = None
        self._when_released = None

    @property
    def is_pressed(self):
        """Get or set the button state as a boolean value.

        :rtype: bool
        """
        return self._is_pressed

    @is_pressed.setter
    def is_pressed(self, value):
        self._is_pressed = value

    @property
    def when_pressed(self):
        """Get or set the 'when pressed' button state callback function. When
        set, this callback function will be invoked when this event happens.

        :type callback: Function
        :param callback:
            Callback function to run when a button is pressed.
        """
        return self._when_pressed

    @when_pressed.setter
    def when_pressed(self, callback):
        self._when_pressed = callback

    @property
    def when_released(self):
        """Get or set the 'when released' button state callback function. When
        set, this callback function will be invoked when this event happens.

        :type callback: Function
        :param callback:
            Callback function to run when a button is released.
        """
        return self._when_released

    @when_released.setter
    def when_released(self, callback):
        self._when_released = callback



import math
from collections import OrderedDict

import numpy as np

FACIAL_LANDMARKS_68_IDXS = OrderedDict(
    [
        ("mouth", (48, 68)),
        ("inner_mouth", (60, 68)),
        ("right_eyebrow", (17, 22)),
        ("left_eyebrow", (22, 27)),
        ("right_eye", (36, 42)),
        ("left_eye", (42, 48)),
        ("nose", (27, 36)),
        ("jaw", (0, 17)),
    ]
)


FACIAL_LANDMARKS_5_IDXS = OrderedDict(
    [("right_eye", (2, 4)), ("left_eye", (0, 2)), ("nose", 4)]
)


def get_face_angle(face_features):
    """Returns angle in degrees of face from dlib face features.

    :param face_features: dlib face features (either 5 landmark or 68 landmark version)
    :return: angle of face in degrees
    """
    if len(face_features) == 68:
        # https://pyimagesearch.com/wp-content/uploads/2017/04/facial_landmarks_68markup.jpg
        # note: array is 0-indexed, image annotations are 1-indexed

        left_eye_start, left_eye_end = FACIAL_LANDMARKS_68_IDXS["left_eye"]
        right_eye_start, right_eye_end = FACIAL_LANDMARKS_68_IDXS["right_eye"]
        left_eyebrow_start, left_eyebrow_end = FACIAL_LANDMARKS_68_IDXS["left_eyebrow"]
        right_eyebrow_start, right_eyebrow_end = FACIAL_LANDMARKS_68_IDXS[
            "right_eyebrow"
        ]
        jaw_start, jaw_end = FACIAL_LANDMARKS_68_IDXS["jaw"]

        left_eye = face_features[left_eye_start:left_eye_end]
        right_eye = face_features[right_eye_start:right_eye_end]

        left_eyebrow = face_features[left_eyebrow_start:left_eyebrow_end]
        right_eyebrow = face_features[right_eyebrow_start:right_eyebrow_end]

        left_jaw = face_features[(jaw_end + 1) // 2 : jaw_end]
        right_jaw = face_features[jaw_start : (jaw_end - 1) // 2]

        left_mouth = np.take(face_features, (52, 53, 54, 55, 56, 63, 64, 65), axis=0)
        right_mouth = np.take(face_features, (48, 49, 50, 58, 59, 60, 61, 67), axis=0)

        all_left_points = np.vstack((left_eye, left_eyebrow, left_jaw, left_mouth))
        all_right_points = np.vstack((right_eye, right_eyebrow, right_jaw, right_mouth))

    elif len(face_features) == 5:
        left_eye_start, left_eye_end = FACIAL_LANDMARKS_5_IDXS["left_eye"]
        right_eye_start, right_eye_end = FACIAL_LANDMARKS_5_IDXS["right_eye"]
        all_left_points = face_features[left_eye_start:left_eye_end]
        all_right_points = face_features[right_eye_start:right_eye_end]

    else:
        raise ValueError("dlib face features not recognised, please try again")

    left_centroid = np.average(all_left_points, axis=0)
    right_centroid = np.average(all_right_points, axis=0)

    position_diff = left_centroid - right_centroid

    x_diff, y_diff = position_diff

    angle = -math.degrees(math.atan(y_diff / x_diff))

    return round(angle, 1)


def pupil_distance(face_features):
    """
    :param face_features: dlib face features (either 5 landmark or 68 landmark version)
    :return: Scalar distance between left eye center and right eye center
    :rtype: float
    """
    return round(
        np.linalg.norm(
            np.asarray(left_eye_center(face_features))
            - np.asarray(right_eye_center(face_features))
        ),
        1,
    )


def left_eye_center(face_features):
    """Left eye is assumed to be on the right of the face in the image frame
    (i.e. from camera's point of view).

    :param face_features: dlib face features (either 5 landmark or 68 landmark version)
    :return: (x, y) position of left eye
    :rtype: tuple
    """
    return feature_center(eye_landmarks(face_features, "left_eye"))


def right_eye_center(face_features):
    """Right eye is assumed to be on the left of the face in the image frame
    (i.e. from camera's point of view).

    :param face_features: dlib face features (either 5 landmark or 68 landmark version)
    :return: (x, y) position of right eye
    :rtype: tuple
    """
    return feature_center(eye_landmarks(face_features, "right_eye"))


def left_eye_dimensions(face_features):
    """Left eye is assumed to be on the right of the face in the image frame
    (i.e. from camera's point of view).

    :param face_features: dlib face features (either 5 landmark or 68 landmark version)
    :return: (width, height) for 68-landmark version and (width, None) for 5-landmark version
    :rtype: tuple
    """
    return eye_dimensions(face_features, "left_eye")


def right_eye_dimensions(face_features):
    """Right eye is assumed to be on the left of the face in the image frame
    (i.e. from camera's point of view).

    :param face_features: dlib face features (either 5 landmark or 68 landmark version)
    :return: (width, height) for 68-landmark version and (width, None) for 5-landmark version
    :rtype: tuple
    """
    return eye_dimensions(face_features, "right_eye")


def eye_dimensions(face_features, position: str):
    landmarks = eye_landmarks(face_features, position)
    if len(face_features) == 68:
        width = round(np.linalg.norm(landmarks[0] - landmarks[3]), 1)
        height = round(
            (
                np.linalg.norm(landmarks[1] - landmarks[5])
                + np.linalg.norm(landmarks[2] - landmarks[4])
            )
            / 2,
            1,
        )
        return width, height
    elif len(face_features) == 5:
        width = np.linalg.norm(landmarks[0] - landmarks[1])
        return width, None


def eye_landmarks(face_features, position: str):
    eye_start, eye_end = get_landmarks_dict(face_features)[position]
    return face_features[eye_start:eye_end]


def mouth_center(face_features):
    return feature_center(mouth_landmarks(face_features))


def mouth_landmarks(face_features):
    if len(face_features) == 5:
        raise ValueError(
            "Not compatible with 5-landmark version, use 68-landmark version instead."
        )
    mouth_start, mouth_end = FACIAL_LANDMARKS_68_IDXS["mouth"]
    return face_features[mouth_start:mouth_end]


def mouth_dimensions(face_features):
    landmarks = mouth_landmarks(face_features)
    width = round(np.linalg.norm(landmarks[0] - landmarks[6]), 1)
    height = round(np.linalg.norm(landmarks[3] - landmarks[9]), 1)
    return width, height


def nose_bottom(face_features):
    """
    :param face_features: dlib face features (either 5 landmark or 68 landmark version)
    :return: (x, y) position of the bottom of the nose
    :rtype: tuple
    """
    if len(face_features) == 68:
        return tuple(np.take(face_features, 33, axis=0).astype(int))
    elif len(face_features) == 5:
        return tuple(np.take(face_features, 4, axis=0).astype(int))
    else:
        raise ValueError("dlib face features not recognised, please try again")


def feature_center(feature_landmarks):
    return tuple(np.mean(feature_landmarks, axis=0).astype(int))


def get_landmarks_dict(face_features):
    if len(face_features) == 68:
        landmarks_dict = FACIAL_LANDMARKS_68_IDXS
    elif len(face_features) == 5:
        landmarks_dict = FACIAL_LANDMARKS_5_IDXS
    else:
        raise ValueError("dlib face features not recognised, please try again")

    return landmarks_dict



import atexit
from os import getenv

from pitop.common.lock import PTLock
from pitop.common.ptdm import Message, PTDMRequestClient, PTDMSubscribeClient

from .contrib.luma.core.interface.serial import spi
from .contrib.luma.oled.device import sh1106


class OledDeviceController:
    SPI_DEVICE = 0
    SPI_BUS_SPEED_HZ = 8000000
    SPI_TRANSFER_SIZE = 4096

    # TODO: drop 'redraw last image' logic at v1.0.0
    #
    # this is only necessary to support users with SPI0 on device
    # with older SDK version that only supported SPI1
    def __init__(self, redraw_last_image_func):
        self.__redraw_last_image_func = redraw_last_image_func
        self.__spi_bus = self.__get_spi_bus_from_ptdm()
        self.__device = None
        self.lock = PTLock("miniscreen")

        self.__ptdm_subscribe_client = None
        self.__setup_subscribe_client()

        atexit.register(self.__clean_up)

    def __setup_subscribe_client(self):
        def on_spi_bus_changed(parameters):
            self.__spi_bus = int(parameters[0])
            self.reset_device()
            self.__redraw_last_image_func()

        self.__ptdm_subscribe_client = PTDMSubscribeClient()
        self.__ptdm_subscribe_client.initialise(
            {Message.PUB_OLED_SPI_BUS_CHANGED: on_spi_bus_changed}
        )
        self.__ptdm_subscribe_client.start_listening()

    def __clean_up(self):
        try:
            self.__ptdm_subscribe_client.stop_listening()
        except Exception:
            pass

    def __set_controls(self, controlled_by_pi):
        message = Message.from_parts(
            Message.REQ_SET_OLED_CONTROL, [str(int(controlled_by_pi))]
        )

        with PTDMRequestClient() as request_client:
            request_client.send_message(message)

    def __setup_device(self):
        if getenv("PT_MINISCREEN_SYSTEM", "0") != "1":
            self.lock.acquire()
            atexit.register(self.reset_device)

        self.__device = sh1106(
            serial_interface=spi(
                port=self.__spi_bus,
                device=self.SPI_DEVICE,
                bus_speed_hz=self.SPI_BUS_SPEED_HZ,
                transfer_size=self.SPI_TRANSFER_SIZE,
                gpio_DC=17 if self.__spi_bus == 1 else 7,  # Always use CE1
                gpio_RST=None,
            ),
            rotate=0,
        )

    def __get_spi_bus_from_ptdm(self):
        message = Message.from_parts(Message.REQ_GET_OLED_SPI_BUS)

        with PTDMRequestClient() as request_client:
            response = request_client.send_message(message)

        return int(response.parameters[0])

    ##############################
    # Public methods
    ##############################

    def device_is_active(self):
        return self.lock.is_locked()

    def reset_device(self):
        self.__device = None
        if self.device_is_active():
            self.lock.release()

    def get_device(self):
        if self.__device is None:
            self.__setup_device()

        return self.__device

    def set_control_to_pi(self):
        self.__set_controls(controlled_by_pi=True)

    def set_control_to_hub(self):
        self.__set_controls(controlled_by_pi=False)

    @property
    def spi_bus(self):
        return self.__spi_bus

    @spi_bus.setter
    def spi_bus(self, bus):
        """Request SPI bus change from pi-top device manager."""
        assert bus in range(0, 2)

        if self.__spi_bus == bus:
            return

        message = Message.from_parts(Message.REQ_SET_OLED_SPI_BUS, [str(bus)])

        with PTDMRequestClient() as request_client:
            request_client.send_message(message)

        # Wait for publish event from ptdm to update.
        #
        # Internal state of SPI bus is handled via subscribe client.
        # When the subscribe client receives an SPI bus change event,
        # the device manager has finished configuration. The response
        # received here simply shows that it has acknowledged the request.



import atexit

from . import const as luma_const
from . import mixin


class device(mixin.capabilities):
    """Base class for display driver classes.

    .. note::
        Direct use of the :func:`command` and :func:`data` methods are
        discouraged: Screen updates should be effected through the
        :func:`display` method, or preferably with the
        :class:`pitop.miniscreen.oled.core.contrib.luma.core.render.canvas` context manager.
    """

    def __init__(self, const=None, serial_interface=None):
        self._const = const or luma_const.common
        self._serial_interface = serial_interface

        def shutdown_hook():  # pragma: no cover
            try:
                self.cleanup()
            except Exception:
                pass

        atexit.register(shutdown_hook)

    def command(self, *cmd):
        """Sends a command or sequence of commands through to the delegated
        serial interface."""
        self._serial_interface.command(*cmd)

    def data(self, data):
        """Sends a data byte or sequence of data bytes through to the delegated
        serial interface."""
        self._serial_interface.data(data)

    def show(self):
        """Sets the display mode ON, waking the device out of a prior low-power
        sleep mode."""
        self.command(self._const.DISPLAYON)

    def hide(self):
        """Switches the display mode OFF, putting the device in low-power sleep
        mode."""
        self.command(self._const.DISPLAYOFF)

    def contrast(self, level):
        """Switches the display contrast to the desired level, in the range
        0-255. Note that setting the level to a low (or zero) value will not
        necessarily dim the display to nearly off. In other words, this method
        is **NOT** suitable for fade-in/out animation.

        :param level: Desired contrast level in the range of 0-255.
        :type level: int
        """
        assert 0 <= level <= 255
        self.command(self._const.SETCONTRAST, level)

    def cleanup(self):
        """Attempt to switch the device off or put into low power mode (this
        helps prolong the life of the device), clear the screen and close
        resources associated with the underlying serial interface.

        If :py:attr:`persist` is ``True``, the device will not be switched off.

        This is a managed function, which is called when the python processs
        is being shutdown, so shouldn't usually need be called directly in
        application code.
        """
        if not self.persist:
            self.hide()
            self.clear()
        self._serial_interface.cleanup()



import math


class GoalCriteria:
    def __init__(
        self,
        full_speed_distance_error: float = 0.02,
        full_speed_angle_error: float = 3.0,
    ):
        self._full_speed_distance_error = full_speed_distance_error
        self._full_speed_angle_error = math.radians(full_speed_angle_error)

        self._max_distance_error = None
        self._max_angle_error = None
        self._starting_angle_error = None

    def angle(self, angle_error):
        if self._starting_angle_error is None:
            self._starting_angle_error = angle_error
        if abs(angle_error) < self._max_angle_error:
            self._starting_angle_error = None
            return True
        if self._starting_angle_error / (angle_error + 1e-6) < 0:
            # if the starting angle error is a different sign to the current angle error, then robot has overshot goal.
            # Consider goal to be reached to avoid trying to adjust angle at low angular speeds where starting friction
            # prevents precise movement
            self._starting_angle_error = None
            return True
        return False

    def distance(self, distance_error, angle_error):
        if abs(distance_error) < self._max_distance_error:
            return True
        if abs(angle_error) > math.pi / 2:
            # Overshot goal, error will be small so consider goal to be reached.
            # Otherwise robot would have to reverse a tiny amount which would be poor UX.
            # Robot will correct itself on subsequent navigation calls anyway.
            return True
        return False

    def update_linear_speed(self, speed_factor):
        self._max_distance_error = speed_factor * self._full_speed_distance_error

    def update_angular_speed(self, speed_factor):
        self._max_angle_error = speed_factor * self._full_speed_angle_error



import math

from ...simple_pid import PID


class DrivingManager:
    class PIDManager:
        def __init__(self):
            self._Ki = 0.1
            self._Kd = 0.1
            self.distance = PID(
                Kp=0.0,
                Ki=self._Ki,
                Kd=self._Kd,
                setpoint=0.0,
                output_limits=(-1.0, 1.0),
            )
            self.heading = PID(
                Kp=0.7,
                Ki=self._Ki,
                Kd=self._Kd,
                setpoint=0.0,
                output_limits=(-1.0, 1.0),
            )

        def reset(self):
            self.heading.reset()
            self.distance.reset()

        def distance_update(self, deceleration_distance):
            self.distance.Kp = 1.0 / deceleration_distance

        def heading_update(self, deceleration_angle):
            # self.heading.Kp = 1.0 / deceleration_angle
            pass

    def __init__(
        self,
        max_motor_speed,
        max_angular_speed,
        full_speed_deceleration_distance=0.4,
        full_speed_deceleration_angle=120.0,
    ):
        self._max_motor_speed = max_motor_speed  # m/s
        self._max_angular_speed = max_angular_speed  # rad/s
        self._full_speed_deceleration_distance = full_speed_deceleration_distance  # m
        self._full_speed_deceleration_angle = full_speed_deceleration_angle  # degrees

        self.linear_speed_factor = None
        self.angular_speed_factor = None
        self.max_velocity = None
        self.max_angular_velocity = None
        self.deceleration_distance = None
        self.deceleration_angle = None
        self.pid = self.PIDManager()

    def update_linear_speed(self, speed_factor):
        self.linear_speed_factor = speed_factor
        self.max_velocity = self.linear_speed_factor * self._max_motor_speed
        self.deceleration_distance = (
            self.linear_speed_factor * self._full_speed_deceleration_distance
        )
        self.pid.distance_update(deceleration_distance=self.deceleration_distance)

    def update_angular_speed(self, speed_factor):
        self.angular_speed_factor = speed_factor
        self.max_angular_velocity = self.angular_speed_factor * self._max_angular_speed
        self.deceleration_angle = self.angular_speed_factor * math.radians(
            self._full_speed_deceleration_angle
        )
        self.pid.heading_update(deceleration_angle=self.deceleration_angle)

    def get_new_angular_speed(self, angle_error):
        return self.max_angular_velocity * self.pid.heading(angle_error)

    def get_new_linear_speed(self, distance_error):
        return self.max_velocity * self.pid.distance(distance_error)



import logging
import math

import numpy as np

from .core.driving_manager import DrivingManager
from .core.goal_criteria import GoalCriteria
from .core.utils import normalize_angle

logger = logging.getLogger(__name__)


class Navigator:
    MINIMUM_HITS = 2

    def __init__(
        self, max_motor_speed, max_robot_angular_speed, measurement_input_function
    ) -> None:
        if not callable(measurement_input_function):
            raise AttributeError(
                "Argument 'measurement_input_function' must be a function"
            )

        self.measurement_input_function = measurement_input_function
        self.backwards = False
        self._stop_triggered = False

        self.goal_criteria = GoalCriteria()
        self.drive_manager = DrivingManager(
            max_motor_speed=max_motor_speed,
            max_angular_speed=max_robot_angular_speed,
        )

    def navigate(self, position, angle):
        position_generators = []
        if position is not None:
            position_generators.append(self.set_course_heading(position))
            position_generators.append(self.drive_to_position_goal(position))
        if angle is not None:
            position_generators.append(self.rotate_to_angle_goal(math.radians(angle)))

        for generator in position_generators:
            for linear_speed, angular_speed in generator:
                yield linear_speed, angular_speed
            yield 0, 0
            self.drive_manager.pid.reset()

    def reached_position(self, position, angle=None):
        x_goal, y_goal = position
        x, y, theta = self.measurement_input_function()
        x_diff, y_diff = self.__get_position_error(
            x=x, x_goal=x_goal, y=y, y_goal=y_goal
        )
        angle_error = 0
        if angle:
            angle_error = self.__get_angle_error(
                current_angle=theta, target_angle=math.atan2(y_diff, x_diff)
            )
        return self.goal_criteria.distance(
            distance_error=self.__get_distance_error(x_diff=x_diff, y_diff=y_diff),
            angle_error=angle_error,
        )

    def get_errors(self, current_pose, goal_pose):
        x, y, theta = current_pose
        x_goal, y_goal, theta_goal = goal_pose

        if x_goal is None and y_goal is None and theta_goal is None:
            raise Exception("At least one element of 'goal_pose' should have a value.")

        distance_error = 0
        if x_goal is not None and y_goal is not None:
            x_diff, y_diff = self.__get_position_error(
                x=x, x_goal=x_goal, y=y, y_goal=y_goal
            )
            distance_error = self.__get_distance_error(x_diff, y_diff)
            if theta_goal is None:
                theta_goal = math.atan2(y_diff, x_diff)

        angle_error = self.__get_angle_error(
            current_angle=theta, target_angle=theta_goal
        )

        return angle_error, distance_error

    def get_motor_input(self, angle_error=None, distance_error=None):
        linear_speed = 0
        angular_speed = 0
        if distance_error:
            linear_speed = self.drive_manager.get_new_linear_speed(
                distance_error=distance_error
            )
        if angle_error:
            angular_speed = self.drive_manager.get_new_angular_speed(
                angle_error=angle_error
            )
        return linear_speed, angular_speed

    def set_course_heading(self, position):
        logger.debug(f"Navigator.set_course_heading - going to position {position}")
        hits = 0
        while not self._stop_triggered:
            current_pose = self.measurement_input_function()
            angle_error, _ = self.get_errors(current_pose, (*position, None))

            if self.reached_position(position) or self.goal_criteria.angle(angle_error):
                hits += 1
                self.drive_manager.pid.reset()

            if hits == self.MINIMUM_HITS:
                logger.debug("Navigator.set_course_heading - arrived, exiting...")
                break

            yield self.get_motor_input(angle_error=angle_error)

    def drive_to_position_goal(self, position):
        logger.debug(f"Navigator.drive_to_position_goal - going to position {position}")
        while not self._stop_triggered:
            current_pose = self.measurement_input_function()
            angle_error, distance_error = self.get_errors(
                current_pose, (*position, None)
            )

            if self.goal_criteria.distance(
                distance_error=distance_error, angle_error=angle_error
            ):
                logger.debug("Navigator.drive_to_position_goal - arrived, exiting...")
                break

            yield self.get_motor_input(
                angle_error=angle_error, distance_error=distance_error
            )

    def rotate_to_angle_goal(self, theta_goal):
        logger.debug(
            f"Navigator.rotate_to_angle_goal - going to angle {math.degrees(theta_goal)}"
        )
        hits = 0
        while not self._stop_triggered:
            current_pose = self.measurement_input_function()
            angle_error, _ = self.get_errors(current_pose, (None, None, theta_goal))

            if self.goal_criteria.angle(angle_error):
                hits += 1
                self.drive_manager.pid.reset()

            if hits == self.MINIMUM_HITS:
                logger.debug("Navigator.rotate_to_angle_goal - arrived, exiting...")
                break

            yield self.get_motor_input(angle_error=angle_error)

    def __get_angle_error(self, current_angle, target_angle):
        return normalize_angle(current_angle - target_angle)

    def __get_position_error(self, x, x_goal, y, y_goal):
        x_diff = x_goal - x if not self.backwards else x - x_goal
        y_diff = y_goal - y if not self.backwards else y - y_goal
        return x_diff, y_diff

    def __get_distance_error(self, x_diff, y_diff):
        value = np.hypot(x_diff, y_diff)
        if not self.backwards:
            return -value
        return value

    @property
    def angular_speed_factor(self):
        return self.drive_manager.angular_speed_factor

    @angular_speed_factor.setter
    def angular_speed_factor(self, speed_factor):
        if not 0.0 < speed_factor <= 1.0:
            raise ValueError("Value must be in the range 0.0 < speed_factor <= 1.0")
        self.drive_manager.update_angular_speed(speed_factor)
        self.goal_criteria.update_angular_speed(speed_factor)

    @property
    def linear_speed_factor(self):
        return self.drive_manager.linear_speed_factor

    @linear_speed_factor.setter
    def linear_speed_factor(self, speed_factor: float):
        if not 0.0 < speed_factor <= 1.0:
            raise ValueError("Value must be in the range 0.0 < speed_factor <= 1.0")
        self.drive_manager.update_linear_speed(speed_factor)
        self.goal_criteria.update_linear_speed(speed_factor)



class Emotion:
    def __init__(self):
        self._type = None
        self._confidence = 0.0
        self._robot_view = None

    def clear(self):
        self.type = None
        self.confidence = 0.0

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def confidence(self):
        return self._confidence

    @confidence.setter
    def confidence(self, value):
        self._confidence = value

    @property
    def robot_view(self):
        return self._robot_view

    @robot_view.setter
    def robot_view(self, value):
        self._robot_view = value

    @property
    def found(self):
        return self.type is not None



from json import dumps


class Stateful:
    """Represents an object with a particular set of important properties that
    represent its state."""

    def __init__(self, children=[]):
        self.children = children

    @property
    def own_state(self):
        """Representation of an object state that will be used to determine the
        current state of an object."""
        return {}

    def __child_state(self, child_name):
        child = getattr(self, child_name)
        if hasattr(child, "state"):
            return child.state
        return None

    @property
    def state(self):
        """Returns a dictionary with the state of the current object and all of
        its children."""
        state = self.own_state
        for k, v in state.items():
            if callable(v):
                state[k] = v()
        for child in self.children:
            child_state = self.__child_state(child)
            if child_state is None:
                continue
            state[child] = child_state
        return state

    def print_state(self):
        print(dumps(self.state, indent=4))



import logging
from math import ceil

logger = logging.getLogger(__name__)


def split_into_bytes(
    data: int, no_of_bytes: int = -1, little_endian: bool = False, signed: bool = False
):
    """This function will split a given integer into a integer list
    representing the bytes to be split;

    :param data: A integer that represents the value of the bytes to be split
    :param no_of_bytes: The number of bytes the integer to be converted into. If the
     no_of_bytes is too small the method returns None.
    :return: A integer list representing a bytes list version of the given integer
    """

    byteorder_indicator = "little" if little_endian is True else "big"

    if no_of_bytes == -1:
        no_of_bytes = ceil(data / 255)
    try:
        byte_string = data.to_bytes(
            no_of_bytes, byteorder=byteorder_indicator, signed=signed
        )
    except OverflowError as e:
        logger.error(e)
        return None

    return [i for i in bytearray(byte_string)]


def join_bytes(byte_list: list, little_endian: bool = False, signed: bool = False):
    """This function will return an integer representing the `byte_list`;

    :param byte_list: A integer list representing a list of bytes
    :return: A integer representing the `byte_list`
    """
    return int.from_bytes(
        byte_list, byteorder="little" if little_endian is True else "big", signed=signed
    )


def get_bits(bitmask_to_get: int, bitwise_data: int):
    """This function will return the integer value represented by the bits that
    are high in `bitmask_to_get` from the bits that represent the integer
    value, `bitwise_data`.

    :param bitmask_to_get: A integer representing the bits you want to get from `bitwise_data`
    :param bitwise_data: A integer representing the data which to extract the bits out of
    :return: A integer representing bits extracted from `bitwise_data`
    """
    return bitmask_to_get & bitwise_data


def to_bits(int_to_convert: int):
    """This function will convert a integer into a binary string without a
    prefix (e.g. '0b');

    :param int_to_convert: A integer to convert into a binary string
    :return: A binary string representing the given integer
    """
    return bin(int_to_convert)[2:]


def from_bits(bin_str: str):
    """This function will convert a binary string (with or without prefix) into
    a integer;

    :param bin_str: A binary string to convert into integer
    :return: A integer representing the given binary string
    """
    return int(bin_str, 2)


def pad_bits(bin_str: str, no_of_bytes: int):
    """This function will 'left zero pad' `bin_str`, if necessary, to ensure
    that it matches length `num_of_bytes`;

    :param bin_str: A binary string which will be padded with '0's if necessary
    :param no_of_bytes: The number of bytes to determine the byte size of the return value
    :return: A string representing bin_str with or without padding
    """
    difference = (no_of_bytes * 8) - len(bin_str)
    if difference <= 0:
        return bin_str
    else:
        padding = "0" * difference
        return padding + bin_str


def flip_bin_string(bin_str: str):
    """This function will flip(invert) each of the bits in a given binary
    string without prefix. i.e. "0" becomes "1" and vice versa;

    :param bin_str: A binary string without prefix to flip
    :return: A binary data represented as string
    """
    flipped_bin_str = ""
    for bit in bin_str:
        flipped_bin_str += str(1 - int(bit))
    return flipped_bin_str


def flip_bits(bitwise_data: int, no_of_bytes: int = -1):
    """This function will flip(invert) the bits of `bitwise_data` and
    `num_of_bytes` allows you to choose the byte size of the value returned.
    This may have padding based on the value of `no_of_bytes`;

    :param bitwise_data: A integer representing the bits to flip
    :param no_of_bytes: A integer to set the number of bytes of the returned value
    :return: A integer representing `bitwise_data` with its bits flipped.
    """
    bin_string_of_data = to_bits(bitwise_data)

    if no_of_bytes <= 0:
        value_to_flip = bin_string_of_data
    else:
        value_to_flip = pad_bits(bin_string_of_data, no_of_bytes)

    flipped_bin_string = flip_bin_string(value_to_flip)
    return from_bits(flipped_bin_string)


def ignore_bits(bitmask_to_ignore: int, bitwise_data: int):
    """This function will set the bits high in `bitmask_to_ignore` low in
    `bitwise_data`;

    :param bitmask_to_ignore: A integer representing the bits you want to ignore
     from `bitwise_data`
    :param bitwise_data: A integer representing the data from which to ignore bits out of
    :return: A integer representing `bitwise_data` with the `bitmask_to_ignore` bits set
     to low
    """
    byte_size_of_data = ceil(bitwise_data.bit_length() / 8)
    return get_bits(flip_bits(bitmask_to_ignore, byte_size_of_data), bitwise_data)


def set_bits_low(bits_to_set_low: int, bitwise_data: int):
    """This function will set bits low based on `bits_to_set_low` in
    `bitwise_data`;

    :param bits_to_set_low: A integer representing the bits to set low in `bitwise_data`
    :param bitwise_data: A integer representing the data from which to set bits low out of
    :return: A integer representing `bitwise_data` with bits set low based on `bits_to_set_low`
    """
    return ignore_bits(bits_to_set_low, bitwise_data)


def set_bits_high(bits_to_set_high: int, bitwise_data: int):
    """This function will set bits high based on `bits_to_set_high` in
    `bitwise_data`;

    :param bits_to_set_high: A integer representing the bits to set high in `bitwise_data`
    :param bitwise_data: A integer representing the data from which to set bits high out of
    :return: A integer representing `bitwise_data` with bits set high based on `bits_to_set_high`
    """
    return bits_to_set_high | bitwise_data


def bitmask_is_on(bitmask_to_check: int, bitwise_data: int):
    """This function will check if bits are high in `bitwise_data` against bits
    that are high in `bitmask_to_check` and return True or False based on that;

    :param bitmask_to_check: A integer representing the bits to check in `bitwise_data`
    :param bitwise_data: A integer representing the data to check bits in
    :return: A boolean returns True only if the bits high in `bitmask_to_check`
     are high in `bitwise_data`
    """
    return all_bits_are_on([bitmask_to_check], bitwise_data)


def all_bits_are_on(bits_to_check: list, bitwise_data: int):
    """This function will determine if particular bits are high in some bitwise
    data.

    :param bits_to_check: A list of bit values to check for. Each value can either be an
     individual bit representation or it can be a 'summed' value which represents OR'd bits
      (e.g. ['0001 0000', '0000 0010' ] or ['0001 0010'])
    :param bitwise_data: The data to be checked
    :return: Boolean representing whether all the bits in
     `bits_to_check` are high in `bitwise_data`
    """
    summed_bits = 0
    for bit_to_check in bits_to_check:
        summed_bits = summed_bits | bit_to_check
    processed_bits = get_bits(summed_bits, bitwise_data)
    return processed_bits == summed_bits



# coding=utf-8
# pynput
# Copyright (C) 2015-2018 Moses Palmér
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""The keyboard implementation for *Windows*."""

# pylint: disable=C0111
# The documentation is extracted from the base classes

# pylint: disable=R0903
# We implement stubs

import contextlib
import ctypes
import enum
from ctypes import wintypes

import pynput._util.win32_vks as VK
import six
from pynput._util import AbstractListener
from pynput._util.win32 import (
    INPUT,
    KEYBDINPUT,
    INPUT_union,
    KeyTranslator,
    ListenerMixin,
    SendInput,
    SystemHook,
    VkKeyScan,
)

from . import _base


class KeyCode(_base.KeyCode):
    def _parameters(self, is_press):
        """The parameters to pass to ``SendInput`` to generate this key.

        :param bool is_press: Whether to generate a press event.

        :return: all arguments to pass to ``SendInput`` for this key

        :rtype: dict
        """
        if self.vk:
            vk = self.vk
            scan = 0
            flags = 0
        else:
            res = VkKeyScan(self.char)
            if (res >> 8) & 0xFF == 0:
                vk = res & 0xFF
                scan = 0
                flags = 0
            else:
                vk = 0
                scan = ord(self.char)
                flags = KEYBDINPUT.UNICODE
        return dict(
            dwFlags=flags | (KEYBDINPUT.KEYUP if not is_press else 0),
            wVk=vk,
            wScan=scan,
        )


class Key(enum.Enum):
    alt = KeyCode.from_vk(VK.MENU)
    alt_l = KeyCode.from_vk(VK.LMENU)
    alt_r = KeyCode.from_vk(VK.RMENU)
    alt_gr = KeyCode.from_vk(VK.RMENU)
    backspace = KeyCode.from_vk(VK.BACK)
    caps_lock = KeyCode.from_vk(VK.CAPITAL)
    cmd = KeyCode.from_vk(VK.LWIN)
    cmd_l = KeyCode.from_vk(VK.LWIN)
    cmd_r = KeyCode.from_vk(VK.RWIN)
    ctrl = KeyCode.from_vk(VK.CONTROL)
    ctrl_l = KeyCode.from_vk(VK.LCONTROL)
    ctrl_r = KeyCode.from_vk(VK.RCONTROL)
    delete = KeyCode.from_vk(VK.DELETE)
    down = KeyCode.from_vk(VK.DOWN)
    end = KeyCode.from_vk(VK.END)
    enter = KeyCode.from_vk(VK.RETURN)
    esc = KeyCode.from_vk(VK.ESCAPE)
    f1 = KeyCode.from_vk(VK.F1)
    f2 = KeyCode.from_vk(VK.F2)
    f3 = KeyCode.from_vk(VK.F3)
    f4 = KeyCode.from_vk(VK.F4)
    f5 = KeyCode.from_vk(VK.F5)
    f6 = KeyCode.from_vk(VK.F6)
    f7 = KeyCode.from_vk(VK.F7)
    f8 = KeyCode.from_vk(VK.F8)
    f9 = KeyCode.from_vk(VK.F9)
    f10 = KeyCode.from_vk(VK.F10)
    f11 = KeyCode.from_vk(VK.F11)
    f12 = KeyCode.from_vk(VK.F12)
    f13 = KeyCode.from_vk(VK.F13)
    f14 = KeyCode.from_vk(VK.F14)
    f15 = KeyCode.from_vk(VK.F15)
    f16 = KeyCode.from_vk(VK.F16)
    f17 = KeyCode.from_vk(VK.F17)
    f18 = KeyCode.from_vk(VK.F18)
    f19 = KeyCode.from_vk(VK.F19)
    f20 = KeyCode.from_vk(VK.F20)
    home = KeyCode.from_vk(VK.HOME)
    left = KeyCode.from_vk(VK.LEFT)
    page_down = KeyCode.from_vk(VK.NEXT)
    page_up = KeyCode.from_vk(VK.PRIOR)
    right = KeyCode.from_vk(VK.RIGHT)
    shift = KeyCode.from_vk(VK.LSHIFT)
    shift_l = KeyCode.from_vk(VK.LSHIFT)
    shift_r = KeyCode.from_vk(VK.RSHIFT)
    space = KeyCode.from_vk(VK.SPACE, char=" ")
    tab = KeyCode.from_vk(VK.TAB)
    up = KeyCode.from_vk(VK.UP)

    insert = KeyCode.from_vk(VK.INSERT)
    menu = KeyCode.from_vk(VK.APPS)
    num_lock = KeyCode.from_vk(VK.NUMLOCK)
    pause = KeyCode.from_vk(VK.PAUSE)
    print_screen = KeyCode.from_vk(VK.SNAPSHOT)
    scroll_lock = KeyCode.from_vk(VK.SCROLL)


class Controller(_base.Controller):
    _KeyCode = KeyCode
    _Key = Key

    def __init__(self, *args, **kwargs):
        super(Controller, self).__init__(*args, **kwargs)

    def _handle(self, key, is_press):
        SendInput(
            1,
            ctypes.byref(
                INPUT(
                    type=INPUT.KEYBOARD,
                    value=INPUT_union(ki=KEYBDINPUT(**key._parameters(is_press))),
                )
            ),
            ctypes.sizeof(INPUT),
        )


class Listener(ListenerMixin, _base.Listener):
    #: The Windows hook ID for low level keyboard events, ``WH_KEYBOARD_LL``
    _EVENTS = 13

    _WM_KEYDOWN = 0x0100
    _WM_KEYUP = 0x0101
    _WM_SYSKEYDOWN = 0x0104
    _WM_SYSKEYUP = 0x0105

    # A bit flag attached to messages indicating that the payload is an actual
    # UTF-16 character code
    _UTF16_FLAG = 0x1000

    # A special virtual key code designating unicode characters
    _VK_PACKET = 0xE7

    #: The messages that correspond to a key press
    _PRESS_MESSAGES = (_WM_KEYDOWN, _WM_SYSKEYDOWN)

    #: The messages that correspond to a key release
    _RELEASE_MESSAGES = (_WM_KEYUP, _WM_SYSKEYUP)

    #: A mapping from keysym to special key
    _SPECIAL_KEYS = {key.value.vk: key for key in Key}

    _HANDLED_EXCEPTIONS = (SystemHook.SuppressException,)

    class _KBDLLHOOKSTRUCT(ctypes.Structure):
        """Contains information about a mouse event passed to a
        ``WH_KEYBOARD_LL`` hook procedure, ``LowLevelKeyboardProc``."""

        _fields_ = [
            ("vkCode", wintypes.DWORD),
            ("scanCode", wintypes.DWORD),
            ("flags", wintypes.DWORD),
            ("time", wintypes.DWORD),
            ("dwExtraInfo", ctypes.c_void_p),
        ]

    #: A pointer to a :class:`KBDLLHOOKSTRUCT`
    _LPKBDLLHOOKSTRUCT = ctypes.POINTER(_KBDLLHOOKSTRUCT)

    def __init__(self, *args, **kwargs):
        super(Listener, self).__init__(*args, **kwargs)
        self._translator = KeyTranslator()
        self._event_filter = self._options.get("event_filter", lambda msg, data: True)

    def _convert(self, code, msg, lpdata):
        if code != SystemHook.HC_ACTION:
            return

        data = ctypes.cast(lpdata, self._LPKBDLLHOOKSTRUCT).contents
        is_packet = data.vkCode == self._VK_PACKET

        # Suppress further propagation of the event if it is filtered
        if self._event_filter(msg, data) is False:
            return None
        elif is_packet:
            return (msg | self._UTF16_FLAG, data.scanCode)
        else:
            return (msg, data.vkCode)

    @AbstractListener._emitter
    def _process(self, wparam, lparam):
        msg = wparam
        vk = lparam

        # If the key has the UTF-16 flag, we treat it as a unicode character,
        # otherwise convert the event to a KeyCode; this may fail, and in that
        # case we pass None
        is_utf16 = msg & self._UTF16_FLAG
        if is_utf16:
            msg = msg ^ self._UTF16_FLAG
            scan = vk
            key = KeyCode.from_char(six.unichr(scan))
        else:
            try:
                key = self._event_to_key(msg, vk)
            except OSError:
                key = None

        if msg in self._PRESS_MESSAGES:
            self.on_press(key)

        elif msg in self._RELEASE_MESSAGES:
            self.on_release(key)

    # pylint: disable=R0201
    @contextlib.contextmanager
    def _receive(self):
        """An empty context manager; we do not need to fake keyboard events."""
        yield

    # pylint: enable=R0201

    def _event_to_key(self, msg, vk):
        """Converts an :class:`_KBDLLHOOKSTRUCT` to a :class:`KeyCode`.

        :param msg: The message received.

        :param vk: The virtual key code to convert.

        :return: a :class:`pynput.keyboard.KeyCode`

        :raises OSError: if the message and data could not be converted
        """
        # We must always call self._translate to keep the keyboard state up to
        # date
        key = KeyCode(**self._translate(vk, msg in self._PRESS_MESSAGES))

        # If the virtual key code corresponds to a Key value, we prefer that
        if vk in self._SPECIAL_KEYS:
            return self._SPECIAL_KEYS[vk]
        else:
            return key

    def _translate(self, vk, is_press):
        """Translates a virtual key code to a parameter list passable to
        :class:`pynput.keyboard.KeyCode`.

        :param int vk: The virtual key code.

        :param bool is_press: Whether this is a press event.

        :return: a paramter list to the :class:`pynput.keyboard.KeyCode`
            constructor
        """
        return self._translator(vk, is_press)



import logging
import signal
from copy import deepcopy
from math import ceil, cos, radians, sin
from os import path
from sys import exit
from threading import Timer
from time import sleep

from serial import Serial, serialutil

from pitop.pulse import configuration

logger = logging.getLogger(__name__)

_initialised = False

_w = 7
_h = 7
_rotation = 0
_brightness = 1.0

_max_freq = 50  # Maximum update speed is 50 times per second
_update_rate = 0.1

_running = False
_show_enabled = True

_gamma_correction_arr = [
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    2,
    3,
    3,
    3,
    3,
    3,
    3,
    3,
    4,
    4,
    4,
    4,
    4,
    5,
    5,
    5,
    5,
    6,
    6,
    6,
    6,
    7,
    7,
    7,
    7,
    8,
    8,
    8,
    9,
    9,
    9,
    10,
    10,
    10,
    11,
    11,
    11,
    12,
    12,
    13,
    13,
    13,
    14,
    14,
    15,
    15,
    16,
    16,
    17,
    17,
    18,
    18,
    19,
    19,
    20,
    20,
    21,
    21,
    22,
    22,
    23,
    24,
    24,
    25,
    25,
    26,
    27,
    27,
    28,
    29,
    29,
    30,
    31,
    32,
    32,
    33,
    34,
    35,
    35,
    36,
    37,
    38,
    39,
    39,
    40,
    41,
    42,
    43,
    44,
    45,
    46,
    47,
    48,
    49,
    50,
    50,
    51,
    52,
    54,
    55,
    56,
    57,
    58,
    59,
    60,
    61,
    62,
    63,
    64,
    66,
    67,
    68,
    69,
    70,
    72,
    73,
    74,
    75,
    77,
    78,
    79,
    81,
    82,
    83,
    85,
    86,
    87,
    89,
    90,
    92,
    93,
    95,
    96,
    98,
    99,
    101,
    102,
    104,
    105,
    107,
    109,
    110,
    112,
    114,
    115,
    117,
    119,
    120,
    122,
    124,
    126,
    127,
    129,
    131,
    133,
    135,
    137,
    138,
    140,
    142,
    144,
    146,
    148,
    150,
    152,
    154,
    156,
    158,
    160,
    162,
    164,
    167,
    169,
    171,
    173,
    175,
    177,
    180,
    182,
    184,
    186,
    189,
    191,
    193,
    196,
    198,
    200,
    203,
    205,
    208,
    210,
    213,
    215,
    218,
    220,
    223,
    225,
    228,
    231,
    233,
    236,
    239,
    241,
    244,
    247,
    249,
    252,
    255,
]

_sync = bytearray(
    [7, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127]
)

_empty = [0, 0, 0]

_empty_map = [
    [_empty, _empty, _empty, _empty, _empty, _empty, _empty],
    [_empty, _empty, _empty, _empty, _empty, _empty, _empty],
    [_empty, _empty, _empty, _empty, _empty, _empty, _empty],
    [_empty, _empty, _empty, _empty, _empty, _empty, _empty],
    [_empty, _empty, _empty, _empty, _empty, _empty, _empty],
    [_empty, _empty, _empty, _empty, _empty, _empty, _empty],
    [_empty, _empty, _empty, _empty, _empty, _empty, _empty],
]

_pixel_map = deepcopy(_empty_map)

#######################
# INTERNAL OPERATIONS #
#######################


def __initialise():
    """INTERNAL.

    Initialise the matrix.
    """

    global _initialised
    global _serial_device
    global _pixel_map

    if not _initialised:
        if configuration.mcu_enabled():
            if not path.exists("/dev/serial0"):
                err_str = "Could not find serial port - are you sure it's enabled?"
                raise serialutil.SerialException(err_str)

            logger.debug("Opening serial port...")

            _serial_device = Serial("/dev/serial0", baudrate=250000, timeout=2)

            if _serial_device.isOpen():
                logger.debug("OK.")
            else:
                logger.info("Error: Failed to open serial port!")
                exit()

            _initialised = True
        else:
            logger.error("Error: pi-topPULSE not initialised by pi-topd")


def __signal_handler(signal, frame):
    """INTERNAL.

    Handles signals from the OS to exit.
    """

    logger.info("\nQuitting...")

    stop()
    off()
    exit(0)


def __get_avg_color():
    """INTERNAL.

    Get the average color of the matrix.
    """

    total_rgb = [0, 0, 0]
    avg_rgb = [0, 0, 0]

    for x in range(_w):
        for y in range(_h):
            for c in range(3):
                total_rgb[c] = total_rgb[c] + _pixel_map[x][y][c]

    for i, val in enumerate(total_rgb):
        avg_rgb[i] = int(round(val / (_w * _h)))

    return avg_rgb


def __write(data):
    """INTERNAL.

    Write data to the matrix.
    """

    logger.debug(
        "{s0:<4}{s1:<4}{s2:<4}{s3:<4}{s4:<4}{s5:<4}{s6:<4}{s7:<4}{s8:<4}{s9:<4}{s10:<4}".format(
            s0=data[0],
            s1=data[1],
            s2=data[2],
            s3=data[3],
            s4=data[4],
            s5=data[5],
            s6=data[6],
            s7=data[7],
            s8=data[8],
            s9=data[9],
            s10=data[10],
        )
    )
    _serial_device.write(data)
    sleep(0.002)


def __get_gamma_corrected_value(original_value):
    """INTERNAL.

    Converts a brightness value from 0-255 to the value that produces an
    approximately linear scaling to the human eye.
    """

    return _gamma_correction_arr[original_value]


def __scale_pixel_to_brightness(original_value):
    """INTERNAL.

    Multiplies intended brightness of a pixel by brightness scaling
    factor to generate an adjusted value.
    """

    unrounded_new_brightness = original_value * _brightness
    rounded_new_brightness = round(unrounded_new_brightness)
    int_new_brightness = int(rounded_new_brightness)

    return int_new_brightness


def __get_rotated_pixel_map():
    """INTERNAL.

    Get a rotated copy of the current in-memory pixel map.
    """

    rotated_pixel_map = deepcopy(_pixel_map)

    # Some fancy maths to rotate pixel map so that
    # 0,0 (x,y) - with rotation 0 - is the bottom left LED
    scaled_rotation = int(_rotation / 90)
    adjusted_scaled_rotation = scaled_rotation + 1
    modulo_adjusted_scaled_rotation = adjusted_scaled_rotation % 4
    count = (6 - modulo_adjusted_scaled_rotation) % 4

    for x in range(count):
        rotated_pixel_map = list(zip(*rotated_pixel_map[::-1]))

    return rotated_pixel_map


def __brightness_correct(original_value):
    """INTERNAL.

    Correct a single color for brightness.
    """

    brightness_scaled = __scale_pixel_to_brightness(original_value)
    new_value = __get_gamma_corrected_value(brightness_scaled)

    return new_value


def __adjust_r_g_b_for_brightness_correction(r, g, b):
    """INTERNAL.

    Correct LED for brightness.
    """

    r = __brightness_correct(r)
    g = __brightness_correct(g)
    b = __brightness_correct(b)

    return r, g, b


def __sync_with_device():
    """INTERNAL.

    Send the sync frame to tell the device that LED data is expected.
    """

    __initialise()
    logger.debug("Sync data:")
    __write(_sync)


def __rgb_to_bytes_to_send(rgb):
    """INTERNAL.

    Format the LED data in the device-specific layout.
    """

    # Create three 5-bit color vals, splitting the green bits
    # into two parts (hardware spec):
    # |XX|G0|G1|R0|R1|R2|R3|R4|
    # |G2|G3|G4|B0|B1|B2|B3|B4|

    r = rgb[0]
    g = rgb[1]
    b = rgb[2]

    byte0 = (r >> 3) & 0x1F
    byte1 = (b >> 3) & 0x1F
    grnb0 = (g >> 1) & 0x60
    grnb1 = (g << 2) & 0xE0

    byte0 = (byte0 | grnb0) & 0xFF
    byte1 = (byte1 | grnb1) & 0xFF

    return byte0, byte1


def __timer_method():
    """INTERNAL.

    Run by the timer on each tick.
    """

    global _running
    global _update_rate

    while _running:
        show()
        sleep(_update_rate)


def __flip(direction):
    """INTERNAL.

    Flip the pixel map.
    """

    global _pixel_map

    flipped_pixel_map = deepcopy(_pixel_map)
    for x in range(_w):
        for y in range(_h):
            if direction == "h":
                flipped_pixel_map[x][y] = _pixel_map[(_w - 1) - x][y]
            elif direction == "v":
                flipped_pixel_map[x][y] = _pixel_map[x][(_h - 1) - y]
            else:
                err = "Flip direction must be [h]orizontal or [v]ertical only"
                raise ValueError(err)

    _pixel_map = flipped_pixel_map


def __set_show_state(enabled):
    """INTERNAL."""

    global _show_enabled

    _show_enabled = enabled

    if not _show_enabled:
        _temp_disable_t.start()


def __enable_show_state():
    """INTERNAL."""

    __set_show_state(True)


def __disable_show_state():
    """INTERNAL."""

    __set_show_state(True)


#######################
# EXTERNAL OPERATIONS #
#######################


def set_debug_print_state(debug_enable):
    """Enable/disable debug prints."""

    global _debug
    _debug = debug_enable


def brightness(new_brightness):
    """Set the display brightness between 0.0 and 1.0.

    :param new_brightness: Brightness from 0.0 to 1.0 (default 1.0)
    """

    global _brightness

    if new_brightness > 1 or new_brightness < 0:
        raise ValueError("Brightness level must be between 0 and 1")
    _brightness = new_brightness


def get_brightness():
    """Get the display brightness value.

    Returns a float between 0.0 and 1.0.
    """

    return _brightness


def rotation(new_rotation=0):
    """Set the display rotation.

    :param new_rotation: Specify the rotation in degrees: 0, 90, 180 or 270
    """

    global _rotation

    if new_rotation in [0, 90, 180, 270]:
        _rotation = new_rotation
        return True
    else:
        raise ValueError("Rotation: 0, 90, 180 or 270 degrees only")


def flip_h():
    """Flips the grid horizontally."""

    __flip("h")


def flip_v():
    """Flips the grid vertically."""

    __flip("v")


def get_shape():
    """Returns the shape (width, height) of the display."""

    return (_w, _h)


def get_pixel(x, y):
    """Get the RGB value of a single pixel.

    :param x: Horizontal position from 0 to 7
    :param y: Veritcal position from 0 to 7
    """

    global _pixel_map

    return _pixel_map[y][x]


def set_pixel(x, y, r, g, b):
    """Set a single pixel to RGB color.

    :param x: Horizontal position from 0 to 7
    :param y: Veritcal position from 0 to 7
    :param r: Amount of red from 0 to 255
    :param g: Amount of green from 0 to 255
    :param b: Amount of blue from 0 to 255
    """

    global _pixel_map

    new_r, new_g, new_b = __adjust_r_g_b_for_brightness_correction(r, g, b)
    _pixel_map[y][x] = [new_r, new_g, new_b]


def set_all(r, g, b):
    """Set all pixels to a specific color."""

    global _pixel_map

    for x in range(_w):
        for y in range(_h):
            new_r, new_g, new_b = __adjust_r_g_b_for_brightness_correction(r, g, b)
            _pixel_map[x][y][0] = new_r
            _pixel_map[x][y][1] = new_g
            _pixel_map[x][y][2] = new_b


def show():
    """Update pi-topPULSE with the contents of the display buffer."""

    global _pixel_map
    global _rotation
    global _show_enabled

    wait_counter = 0

    attempt_to_show_early = not _show_enabled
    if attempt_to_show_early:
        logger.info("Can't update pi-topPULSE LEDs more than 50/s. Waiting...")

    pause_length = 0.001

    # Scale wait time to _max_freq
    wait_counter_length = ceil(float(1 / float(_max_freq * pause_length)))

    while not _show_enabled:
        if wait_counter >= wait_counter_length:
            # Timer hasn't reset for some reason - force override
            __enable_show_state()
            break
        else:
            sleep(pause_length)
            wait_counter = wait_counter + 1

    if attempt_to_show_early:
        logger.debug("pi-topPULSE LEDs re-enabled.")

    __sync_with_device()

    rotated_pixel_map = __get_rotated_pixel_map()
    avg_rgb = __get_avg_color()

    __initialise()

    logger.debug("LED data:")
    # For each col
    for x in range(_w):
        # Write col to LED matrix
        # Start with col no., so LED matrix knows which one it belongs to
        pixel_map_buffer = chr(x)
        # Get col's frame buffer, iterating over each pixel
        for y in range(_h + 1):
            if y == _h:
                # Ambient lighting bytes
                byte0, byte1 = __rgb_to_bytes_to_send(avg_rgb)
            else:
                byte0, byte1 = __rgb_to_bytes_to_send(rotated_pixel_map[x][y])

            pixel_map_buffer += chr(byte0)
            pixel_map_buffer += chr(byte1)

        # Write col to LED matrix
        arr = bytearray(pixel_map_buffer, "Latin_1")
        __write(arr)

        # Prevent another write if it's too fast
        __disable_show_state()


def clear():
    """Clear the buffer."""

    global _pixel_map

    _pixel_map = deepcopy(_empty_map)


def off():
    """Clear the buffer and immediately update pi-topPULSE."""

    clear()
    show()


def run_tests():
    """Runs a series of tests to check the LED board is working as expected."""

    off()

    # ------------------------------
    # Pixels
    # ------------------------------

    counter = 0

    for r in range(4):
        rotation(90 * r)
        for x in range(_w):
            for y in range(_h):
                rad = radians((float(counter) / (4 * _w * _h)) * 360)

                r = int((sin(rad) * 127) + 127)
                g = int((cos(rad) * 127) + 127)
                b = 255 - int((sin(rad) * 127) + 127)

                set_pixel(x, y, r, g, b)
                show()
                sleep(0.05)
                counter = counter + 1
        off()

    sleep(0.2)

    # ------------------------------
    # Rows and rotation
    # ------------------------------

    for r in range(4):
        rotation(90 * r)
        for c in range(3):
            for x in range(_w):
                for y in range(_h):
                    set_pixel(
                        x,
                        y,
                        255 if c == 0 else 0,
                        255 if c == 1 else 0,
                        255 if c == 2 else 0,
                    )

                show()
                sleep(0.05)

    off()
    sleep(0.2)

    # ------------------------------
    # Brightness
    # ------------------------------

    for b in range(100):
        brightness(float(b) / 100)
        set_all(255, 255, 255)
        show()
        sleep(0.01)

    for b in range(100):
        brightness(1 - (float(b) / 100))
        set_all(255, 255, 255)
        show()
        sleep(0.01)

    off()
    brightness(1.0)

    sleep(0.2)

    # ------------------------------
    # Flipping
    # ------------------------------

    for x in range(int(_w / 2)):
        for y in range(int(_h / 2)):
            set_pixel(x, y, 255, 255, 255)

    set_pixel(int(_w / 4), int(_h / 4), 0, 255, 0)

    show()
    sleep(0.5)

    for f in range(4):
        for x in range(2):
            if x == 0:
                flip_h()
            else:
                flip_v()
            show()
            sleep(0.5)

    off()
    sleep(0.2)

    # ------------------------------
    # Conway - auto refresh
    # ------------------------------

    start(0.1)

    life_map = [
        [0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0],
        [1, 1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0],
    ]

    for r in range(40):
        temp_map = deepcopy(life_map)
        for x in range(_w):
            for y in range(_h):

                current_cell = temp_map[x][y]
                neighbours = 0
                neighbours = neighbours + temp_map[(x - 1) % _w][(y - 1) % _h]
                neighbours = neighbours + temp_map[(x - 1) % _w][(y - 0) % _h]
                neighbours = neighbours + temp_map[(x - 1) % _w][(y + 1) % _h]
                neighbours = neighbours + temp_map[(x - 0) % _w][(y - 1) % _h]
                neighbours = neighbours + temp_map[(x - 0) % _w][(y + 1) % _h]
                neighbours = neighbours + temp_map[(x + 1) % _w][(y - 1) % _h]
                neighbours = neighbours + temp_map[(x + 1) % _w][(y - 0) % _h]
                neighbours = neighbours + temp_map[(x + 1) % _w][(y + 1) % _h]

                if current_cell == 1 and (neighbours < 2 or neighbours > 3):
                    life_map[x][y] = 0

                if current_cell == 0 and neighbours == 3:
                    life_map[x][y] = 1

        for x in range(_w):
            for y in range(_h):
                if life_map[x][y] == 1:
                    set_pixel(x, y, 255, 255, 0)
                else:
                    set_pixel(x, y, 0, 128, 0)

        sleep(0.1)

    stop()
    off()


def start(new_update_rate=0.1):
    """Starts a timer to automatically refresh the LEDs."""

    global _update_rate
    global _running
    global _auto_refresh_timer

    if new_update_rate < (1 / _max_freq):
        _update_rate = 1 / _max_freq
    else:
        _update_rate = new_update_rate

    _running = True
    _auto_refresh_timer.start()


def stop():
    """Stops the timer that automatically refreshes the LEDs."""

    global _running
    global _auto_refresh_timer

    _running = False
    _auto_refresh_timer.cancel()


##################
# INITIALISATION #
##################

_signal = signal.signal(signal.SIGINT, __signal_handler)
_auto_refresh_timer = Timer(_update_rate, __timer_method)
_temp_disable_t = Timer(_max_freq, __enable_show_state)

clear()



from os import environ

import numpy as np

from pitop.common.current_session_info import get_first_display

if not environ.get("DISPLAY"):
    environ["DISPLAY"] = str(get_first_display())


def plot_ellipsoid(
    center, radii, rotation, ax=None, plotAxes=False, cageColor="b", cageAlpha=0.2
):
    """Plot an ellipsoid."""
    u = np.linspace(0.0, 2.0 * np.pi, 100)
    v = np.linspace(0.0, np.pi, 100)

    # cartesian coordinates that correspond to the spherical angles:
    x = radii[0] * np.outer(np.cos(u), np.sin(v))
    y = radii[1] * np.outer(np.sin(u), np.sin(v))
    z = radii[2] * np.outer(np.ones_like(u), np.cos(v))
    # rotate accordingly
    for i in range(len(x)):
        for j in range(len(x)):
            [x[i, j], y[i, j], z[i, j]] = (
                np.dot([x[i, j], y[i, j], z[i, j]], rotation) + center
            )

    if plotAxes:
        # make some purdy axes
        axes = np.array(
            [[radii[0], 0.0, 0.0], [0.0, radii[1], 0.0], [0.0, 0.0, radii[2]]]
        )
        # rotate accordingly
        for i in range(len(axes)):
            axes[i] = np.dot(axes[i], rotation)

        # plot axes
        for p in axes:
            X3 = np.linspace(-p[0], p[0], 100) + center[0]
            Y3 = np.linspace(-p[1], p[1], 100) + center[1]
            Z3 = np.linspace(-p[2], p[2], 100) + center[2]
            ax.plot(X3, Y3, Z3, color=cageColor)

    # plot ellipsoid
    if ax:
        ax.plot_wireframe(
            x, y, z, rstride=4, cstride=4, color=cageColor, alpha=cageAlpha
        )


def least_squares_ellipsoid_fit(magX, magY, magZ):
    # ax^2 + by^2 + cz^2 +2fyz + 2gxz + 2hxy + px + qy + rz + d = 0
    # x_T.M.x + x_T.n + d = 0

    a1 = magX ** 2
    a2 = magY ** 2
    a3 = magZ ** 2
    a4 = 2 * np.multiply(magY, magZ)
    a5 = 2 * np.multiply(magX, magZ)
    a6 = 2 * np.multiply(magX, magY)
    a7 = 2 * magX
    a8 = 2 * magY
    a9 = 2 * magZ
    a10 = np.ones(len(magX)).T
    D = np.array([a1, a2, a3, a4, a5, a6, a7, a8, a9, a10])

    # Eqn 7, k = 4
    C1 = np.array(
        [
            [-1, 1, 1, 0, 0, 0],
            [1, -1, 1, 0, 0, 0],
            [1, 1, -1, 0, 0, 0],
            [0, 0, 0, -4, 0, 0],
            [0, 0, 0, 0, -4, 0],
            [0, 0, 0, 0, 0, -4],
        ]
    )

    # Eqn 11
    S = np.matmul(D, D.T)
    S11 = S[:6, :6]
    S12 = S[:6, 6:]
    S21 = S[6:, :6]
    S22 = S[6:, 6:]

    # Eqn 15, find eigenvalue and vector
    # Since S is symmetric, S12.T = S21
    tmp = np.matmul(
        np.linalg.inv(C1), S11 - np.matmul(S12, np.matmul(np.linalg.inv(S22), S21))
    )
    eigenValue, eigenVector = np.linalg.eig(tmp)
    u1 = eigenVector[:, np.argmax(eigenValue)]

    # Eqn 13 solution
    u2 = np.matmul(-np.matmul(np.linalg.inv(S22), S21), u1)

    # Total solution
    u = np.concatenate([u1, u2]).T

    M = np.array([[u[0], u[5], u[4]], [u[5], u[1], u[3]], [u[4], u[3], u[2]]])

    n = np.array([[u[6]], [u[7]], [u[8]]])

    d = u[9]

    return M, n, d


def get_ellipsoid_geometric_params(M, n, d):
    a = M[0, 0]
    b = M[1, 1]
    c = M[2, 2]
    f = M[2, 1]
    g = M[0, 2]
    h = M[0, 1]

    p = n[0][0]
    q = n[1][0]
    r = n[2][0]

    Q = np.array([[a, h, g, p], [h, b, f, q], [g, f, c, r], [p, q, r, d]])

    Minv = np.linalg.inv(M)
    b = -np.dot(Minv, n)
    center = b.T[0]

    Tofs = np.eye(4)
    Tofs[3, 0:3] = center

    R = np.dot(Tofs, np.dot(Q, Tofs.T))

    R3 = R[0:3, 0:3]

    s1 = -R[3, 3]
    R3S = R3 / s1
    eigenvalues, eigenvectors = np.linalg.eig(R3S)

    rotation_matrix = np.linalg.inv(
        eigenvectors
    )  # inverse is actually the transpose here

    eigen_reciprocal = 1.0 / np.abs(eigenvalues)
    radii = np.sqrt(eigen_reciprocal)

    return center, radii, rotation_matrix



from os.path import isdir, isfile, join
from time import sleep, strftime

from pitop.common.formatting import is_url

from .cli_base import CliBaseClass, PitopCliInvalidArgument


class OledCLI(CliBaseClass):
    parser_help = "pi-top OLED quick text"
    cli_name = "oled"

    def __init__(self, args) -> None:
        self.args = args
        self.validate_args()

    def validate_args(self) -> None:
        if self.args.oled_subcommand is None:
            raise PitopCliInvalidArgument

    def run(self) -> int:
        def path_to_stored_file(arg_path):
            file_path = ""
            if isfile(arg_path):
                file_path = arg_path
            elif isdir(arg_path):
                file_path = join(arg_path, f"output_{strftime('%Y-%m-%d-%H-%M-%S')}")
            else:
                raise Exception(f"Invalid path: {arg_path}")
            return file_path

        try:
            from pitop.miniscreen import Miniscreen

            if self.args.oled_subcommand == "display":
                oled = Miniscreen()

                if self.args.force:
                    oled.set_control_to_pi()

                try:
                    print("Press Ctrl + C to exit.")

                    skip_timeout = False
                    if isfile(self.args.text) or is_url(self.args.text):
                        oled.play_animated_image_file(self.args.text)
                    else:
                        oled.display_multiline_text(
                            self.args.text, font_size=self.args.font_size
                        )

                    if not skip_timeout:
                        sleep(self.args.timeout)

                except KeyboardInterrupt:
                    pass

            elif self.args.oled_subcommand == "spi":
                oled = Miniscreen()

                if self.args.spi_bus is not None:
                    oled.spi_bus = self.args.spi_bus
                else:
                    print(oled.spi_bus)

            elif self.args.oled_subcommand == "capture":
                if self.args.capture_subcommand == "save":
                    file_path = path_to_stored_file(self.args.path)
                    extension = "PNG"
                    print(f"Saving capture to {file_path}.{extension}")
                    oled.image.save(f"{file_path}.{extension}")
                elif self.args.capture_subcommand == "start":
                    file_path = path_to_stored_file(self.args.path)
                    extension = "MP4"
                    print(f"Saving video capture to {file_path}.{extension}")
                    # TODO: configure video capture from OLED
                elif self.args.capture_subcommand == "stop":
                    print("Stopping video capture")
                    # TODO: configure video capture from OLED

            return 0
        except Exception as e:
            print(f"Error on pitop-oled.run: {e}")
            return 1

    @classmethod
    def add_parser_arguments(cls, parser) -> None:
        subparser = parser.add_subparsers(
            title="OLED screen utilities",
            description="Set of utilities to use pi-top [4]'s OLED screen",
            dest="oled_subcommand",
        )

        # "capture" arguments
        parser_capture = subparser.add_parser(
            "capture", help="Capture images or videos of the OLED content"
        )
        parser_capture.add_argument(
            "--force",
            "-f",
            help="Force the hub to give control of the OLED to the Pi",
            action="store_true",
        )
        capture_subparser = parser_capture.add_subparsers(
            title="Capture images or videos of the OLED content",
            description="description",
            dest="capture_subcommand",
        )
        # "capture save" arguments
        save_parser = capture_subparser.add_parser(
            "save",
            help="Store an image in the given path with the current content of the OLED",
        )
        save_parser.add_argument(
            "--path",
            "-p",
            type=str,
            help="Path where snap will be stored. Defaults to /tmp/",
            default="/tmp/",
        )

        # "capture start" arguments
        start_parser = capture_subparser.add_parser(
            "start",
            help="Start recording OLED screen",
        )
        start_parser.add_argument(
            "--path",
            "-p",
            type=str,
            help="Path where video will be stored. Defaults to /tmp/",
            default="/tmp/",
        )

        capture_subparser.add_parser(
            "stop",
            help="Stop recording OLED screen",
        )
        # "display" arguments
        parser_display = subparser.add_parser(
            "display", help="Display text and images on the OLED"
        )
        parser_display.add_argument(
            "--force",
            "-f",
            help="Force the hub to give control of the OLED to the Pi",
            action="store_true",
        )
        parser_display.add_argument(
            "--timeout",
            "-t",
            type=int,
            help="Set the timeout in seconds",
            default=10,
        )
        parser_display.add_argument(
            "--font-size",
            "-s",
            type=int,
            help="Set the font size",
            default=20,
        )
        parser_display.add_argument(
            "--loop",
            "-l",
            type=int,
            help="How many times the animated image should be looped",
            default=1,
        )
        parser_display.add_argument(
            "text",
            help="Set the text to write to screen",
        )

        # "spi" arguments
        parser_spi = subparser.add_parser(
            "spi", help="Set SPI bus that is used by OLED"
        )
        parser_spi.add_argument(
            "spi_bus",
            help="SPI bus to be used by OLED. Valid options: {0, 1}",
            type=int,
            choices=[0, 1],
            nargs="?",
        )


def main():
    from .deprecated_cli_runner import run

    run(OledCLI)


if __name__ == "__main__":
    main()



from pitop.core.mixins import Recreatable, Stateful
from pitop.pma.common.utils import Port

from .ultrasonic_sensor_base import UltrasonicSensorMCU, UltrasonicSensorRPI

valid_analog_ports = ["A1", "A3"]


class UltrasonicSensor(Stateful, Recreatable):
    def __init__(
        self,
        port_name,
        queue_len=5,
        max_distance=3,
        threshold_distance=0.3,
        partial=False,
        name="ultrasonic",
    ):

        assert port_name in Port
        self._pma_port = port_name
        self.name = name

        if port_name in valid_analog_ports:
            self.__ultrasonic_device = UltrasonicSensorMCU(
                port_name=port_name,
                queue_len=queue_len,
                max_distance=max_distance,
                threshold_distance=threshold_distance,
                partial=partial,
                name=name,
            )
        else:
            self.__ultrasonic_device = UltrasonicSensorRPI(
                port_name=port_name,
                queue_len=queue_len,
                max_distance=max_distance,
                threshold_distance=threshold_distance,
                partial=partial,
                name=name,
            )

        self._in_range_function = None
        self._out_of_range_function = None

        Stateful.__init__(self)
        Recreatable.__init__(
            self,
            {
                "port_name": port_name,
                "queue_len": queue_len,
                "partial": partial,
                "name": self.name,
                "max_distance": lambda: self.max_distance,
                "threshold_distance": lambda: self.threshold_distance,
            },
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def own_state(self):
        return {
            "distance": self.distance,
        }

    @property
    def max_distance(self):
        """The maximum distance that the sensor will measure in meters.

        This value is specified in the constructor and is used to
        provide the scaling for the :attr:`~SmoothedInputDevice.value`
        attribute. When :attr:`distance` is equal to
        :attr:`max_distance`, :attr:`~SmoothedInputDevice.value` will be
        1.
        """
        return self.__ultrasonic_device.max_distance

    @max_distance.setter
    def max_distance(self, value):
        self.__ultrasonic_device.max_distance = value

    @property
    def threshold_distance(self):
        """
        The distance, measured in meters, that will trigger the
        :attr:`when_in_range` and :attr:`when_out_of_range` events when
        crossed. This is simply a meter-scaled variant of the usual
        :attr:`~SmoothedInputDevice.threshold` attribute.
        """
        return self.__ultrasonic_device.threshold_distance

    @threshold_distance.setter
    def threshold_distance(self, value):
        self.__ultrasonic_device.threshold_distance = value

    @property
    def distance(self):
        """Returns the current distance measured by the sensor in meters.

        Note
        that this property will have a value between 0 and
        :attr:`max_distance`.
        """
        return self.__ultrasonic_device.distance

    @property
    def value(self):
        """Returns a value between 0, indicating that something is either
        touching the sensor or is sufficiently near that the sensor can't tell
        the difference, and 1, indicating that something is at or beyond the
        specified *max_distance*."""
        return self.__ultrasonic_device.value

    @property
    def pin(self):
        return self.__ultrasonic_device.pin

    def close(self):
        """Shut down the device and release all associated resources. This
        method can be called on an already closed device without raising an
        exception.

        This method is primarily intended for interactive use at the command
        line. It disables the device and releases its pin(s) for use by another
        device.

        You can attempt to do this simply by deleting an object, but unless
        you've cleaned up all references to the object this may not work (even
        if you've cleaned up all references, there's still no guarantee the
        garbage collector will actually delete the object at that point).  By
        contrast, the close method provides a means of ensuring that the object
        is shut down.

        For example, if you have a buzzer connected to port D0, but then wish
        to attach an LED instead:

            >>> from pitop import Buzzer, LED
            >>> bz = Buzzer("D0")
            >>> bz.on()
            >>> bz.off()
            >>> bz.close()
            >>> led = LED("D0")
            >>> led.blink()

        :class:`Device` descendents can also be used as context managers using
        the :keyword:`with` statement. For example:

            >>> from pitop import Buzzer, LED
            >>> with Buzzer("D0") as bz:
            ...     bz.on()
            ...
            >>> with LED("D0") as led:
            ...     led.on()
            ...
        """
        self.__ultrasonic_device.close()

    @property
    def in_range(self):
        return self.__ultrasonic_device.in_range

    @property
    def when_in_range(self):
        return self.__ultrasonic_device.when_deactivated

    @when_in_range.setter
    def when_in_range(self, function):
        if callable(function):
            self.__ultrasonic_device.when_deactivated = function

    @property
    def when_out_of_range(self):
        return self.__ultrasonic_device.when_activated

    @when_out_of_range.setter
    def when_out_of_range(self, function):
        if callable(function):
            self.__ultrasonic_device.when_activated = function

    def wait_for_in_range(self, timeout=None):
        self.__ultrasonic_device.wait_for_inactive(timeout=timeout)

    def wait_for_out_of_range(self, timeout=None):
        self.__ultrasonic_device.wait_for_active(timeout=timeout)
